import discord

from levelling.dataclass import LevellingMember


class LevelUpPayload:
    """
    This class represents the dispatched payload when a Member levels up

    Attributes
    ----------
    member : LevellingMember
        The Member who levelled up.

        Note this is this packages member, you will need to fetch the
        member object yourself to get a ``discord.Member`` object
    level : int
        The new level of the Member levelling up
    channel : discord.TextChannel
        The channel the message that triggered a level up occurred in
    guild : discord.Guild
        The LevellingGuild for the member levelling up

    Notes
    -----
    This is not initialized by you.
    """

    __slots__ = ("member", "level", "channel", "guild")

    def __init__(
        self, member: LevellingMember, level: int, channel: discord.TextChannel
    ):
        self.member: LevellingMember = member
        self.level: int = level
        self.channel: discord.TextChannel = channel
        self.guild: discord.Guild = self.channel.guild



from ..abc import Cache
from ..dataclass import LevellingGuild, LevellingMember
from ..exceptions import GuildNotFound, MemberNotFound


class Memory(Cache):
    """An in memory cache utilising a dictionary internally"""

    def __init__(self):
        self.cache = {}

    async def get_guild(self, guild_id: int) -> LevellingGuild:
        try:
            return LevellingGuild(id=guild_id, raw_members=self.cache[guild_id])
        except KeyError:
            raise GuildNotFound from None

    async def get_member(self, member_id: int, guild_id: int = None) -> LevellingMember:
        if guild_id:
            try:
                guild: LevellingGuild = await self.get_guild(guild_id=guild_id)
                member: LevellingMember = next(
                    i for i in guild.members if i.id == member_id
                )
                return member
            except GuildNotFound:
                raise MemberNotFound from None
            except KeyError:
                raise MemberNotFound from None
            except StopIteration:
                raise MemberNotFound

        try:
            return self.cache[member_id]
        except KeyError:
            raise MemberNotFound from None

    async def set_member(
        self, member_id: int, data: dict, guild_id: int = None
    ) -> None:
        if guild_id:
            guild = self.cache.get(guild_id)
            if not guild:
                self.cache[guild_id] = {}
            self.cache[guild_id][member_id] = data
        else:
            self.cache[member_id] = data



from typing import Protocol, List

from levelling.dataclass import LevellingMember, LevellingGuild


class Datastore(Protocol):
    """A base interface for datastores"""

    __slots__ = ()

    async def fetch_guild(self, guild_id: int) -> LevellingGuild:
        """Returns the associated guilds data"""
        raise NotImplementedError

    async def fetch_member(
        self, member_id: int, guild_id: int = None
    ) -> LevellingMember:
        """Returns the associated member data

        Where ``guild_id`` is ``None`` implies
        a global levelling storage
        """
        raise NotImplementedError

    async def set_member(
        self, member_id: int, data: dict, guild_id: int = None
    ) -> None:
        """Store data within the cache attached to a member

        Where ``guild_id`` is ``None`` implies
        a global levelling storage
        """
        raise NotImplementedError

    async def fetch_all_members(self, guild_id: int = None) -> List[LevellingMember]:
        """
        Returns all members for a guild, or everyone.

        Parameters
        ----------
        guild_id: int, optional
            The guild to fetch from

        Returns
        -------
        List[LevellingMember]
            A list of members sorted by level
        """
        raise NotImplementedError


class Cache(Protocol):
    """A base interface for cache's"""

    __slots__ = ()

    async def get_guild(self, guild_id: int) -> LevellingGuild:
        """Returns the associated guilds data"""
        raise NotImplementedError

    async def get_member(self, member_id: int, guild_id: int = None) -> LevellingMember:
        """Returns the associated member data

        Where ``guild_id`` is ``None`` implies
        a global levelling storage
        """
        raise NotImplementedError

    async def set_member(
        self, member_id: int, data: dict, guild_id: int = None
    ) -> None:
        """Store data within the cache attached to a member

        Where ``guild_id`` is ``None`` implies
        a global levelling storage
        """
        raise NotImplementedError



__version__ = "0.1.3"

import logging
from collections import namedtuple

from levelling.level import Level
from levelling.payloads import LevelUpPayload

logging.getLogger(__name__).addHandler(logging.NullHandler())
VersionInfo = namedtuple("VersionInfo", "major minor micro releaselevel serial")
version_info = VersionInfo(major=0, minor=1, micro=3, releaselevel="alpha", serial=0)



import math
import random
from functools import lru_cache
from typing import Optional, List

import discord

from levelling.store import Store
from levelling.caches import Memory
from levelling.options import Options
from levelling.dataclass import LevellingMember
from levelling.datastores import Sqlite
from levelling.abc import Cache, Datastore
from levelling.payloads import LevelUpPayload


class Level:
    """The main interaction point for the discord.ext.levelling package"""

    def __init__(
        self,
        bot,
        *,
        cache: Cache = None,
        datastore: Datastore = None,
        options: Options = None
    ):
        """

        Parameters
        ----------
        bot: commands.Bot
            The instance of the bot to use internally
        cache : Cache
            An instance of a class implementing the ``Cache`` interface Protocol
        datastore : Datastore
            An instance of a class implementing the ``Datastore`` interface Protocol
        options : Options
            An instance of the ``Options`` dataclass to show the options to support
        """
        self.bot = bot
        self.options = options or Options()

        cache = cache or Memory()
        datastore = datastore or Sqlite()
        self.store = Store(cache=cache, datastore=datastore)

    async def propagate(self, message: discord.Message) -> Optional[LevelUpPayload]:
        if (self.options.ignore_dms or self.options.per_guild) and not message.guild:
            # Either dm's are ignored or per guild is enabled, so ignore dm's
            # Ignore guilds atm
            return

        if self.options.ignore_bots and message.author.bot:
            return

        guild_id: int = message.guild.id if self.options.per_guild else None
        member: LevellingMember = await self.store.create_or_fetch_member(
            member_id=message.author.id, guild_id=guild_id
        )

        # Get level before xp
        current_level = self.get_level_from_xp(member.xp)

        # Add a random amount of xp with the base
        member.xp += (
            math.floor(random.random() * 10) + self.options.xp_base_increase_amount
        )

        # Get level after xp addition
        new_level = self.get_level_from_xp(member.xp)

        # Update internals
        await self.store.set_member(member)

        if current_level == new_level:
            # Do nothing, didnt level up
            return None

        # Did level up
        return LevelUpPayload(member=member, level=new_level, channel=message.channel)

    async def leaderboard(self, guild_id: int = None) -> List[LevellingMember]:
        guild_id: int = guild_id if self.options.per_guild else None
        members = await self.store.datastore.fetch_all_members(guild_id)
        data = []
        for member in members:
            member.level = self.get_level_from_xp(member.xp)
            data.append(member)

        return sorted(data, key=lambda x: x.level)

    @lru_cache
    def get_level_xp_amount(self, level: int) -> int:
        """
        Returns the amount of xp required
        for the next level

        Parameters
        ----------
        level: int
            The members current level

        Returns
        -------
        int
            Required xp for the next level
        """
        return 5 * (level ** 2) + 50 * level + 100

    @lru_cache
    def get_level_from_xp(self, xp: int) -> int:
        """
        Given a member, extract there level from the
        amount of xp they currently have

        Parameters
        ----------
        xp: int
            The xp amount to get a level for

        Returns
        -------
        int
            The LevellingMember's level
        """
        level = 0
        remaining_xp = xp
        while remaining_xp >= self.get_level_xp_amount(level):
            remaining_xp -= self.get_level_xp_amount(level)
            level += 1
        return level

    @lru_cache
    def get_remaining_xp(self, current_level: int, xp: int) -> int:
        """
        Calculates the required amount of xp left to
        achieve the next level from the current level

        Parameters
        ----------
        current_level : int
            The Members current level
        xp : int
            The members current xp

        Returns
        -------
        int
            The amount of xp required to get to the next level
        """
        xp_to_next_level = 0
        for i in range(current_level):
            xp_to_next_level += self.get_level_xp_amount(current_level)
        return xp - xp_to_next_level

    @staticmethod
    async def from_store(datastore: Datastore):
        """
        Creates and returns a Level instance from
        a given ``Datastore``

        Parameters
        ----------
        datastore : Datastore
            The datastore to restore from

        Returns
        -------
        Level
            A ``Level`` instance built from a given ``Datastore``
        """



from discord import DiscordException


class LevellingError(DiscordException):
    """Base Levelling exception class"""

    pass


class MemberNotFound(LevellingError):
    """Raised internally when a member doesn't exist within the cache"""

    pass


class GuildNotFound(LevellingError):
    """Raised internally when a guild doesn't exist within the cache"""

    pass



from typing import List, Dict, Optional

import attr


@attr.s(slots=True)
class LevellingMember:
    """A generic ``attrs`` dataclass representing a ``discord.LevellingMember``

    Parameters
    ----------
    id : int
        -> ``discord.Member.id``
    xp : int
        The total xp this LevellingMember has
    guild_id : int, optional
        Which guild this LevellingMember is attached to.
        If this is null it means levels are stored
        globally rather then per guild
    """

    id: int = attr.ib(hash=True, eq=True)  # Think member.id
    xp: int = attr.ib(eq=False, default=0)
    guild_id: int = attr.ib(
        default=None,
        validator=attr.validators.optional(attr.validators.instance_of(int)),
        kw_only=True,
        eq=False,
    )
    level: Optional[int] = attr.ib(default=None)

    @xp.validator
    def _validate_positive(self, attr, value):
        if value < 0:
            raise ValueError("XP must be positive")


@attr.s
class LevellingGuild:
    """A generic ``attrs`` dataclass representing a ``discord.LevellingGuild``

    One of ``members`` | ``raw_members`` are **required** to build
    a LevellingGuild, however this is not checked/enforced

    Parameters
    ----------
    id : int
        -> ``discord.Guild.id``
    members: List[LevellingMember], optional
        The members internally associated with this guild

    Other Parameters
    ----------------
    raw_members : dict, optional
        A Dict of all the members in this guild. This
        will get lazily built into ``members`` when required
    """

    id: int = attr.ib(hash=True)  # Think guild.id
    _members: List[LevellingMember] = attr.ib(default=attr.Factory(list))
    _raw_members: Dict = attr.ib(default=attr.Factory(dict))

    @property
    def members(self) -> List[LevellingMember]:
        if not bool(self._members):
            for k, v in self._raw_members.items():
                # k id, v actual data
                self._members.append(LevellingMember(**v))

        return self._members



import functools
import os
import sys
from typing import List

import aiosqlite as aiosqlite

from levelling.abc import Datastore
from levelling.dataclass import LevellingGuild, LevellingMember
from levelling.exceptions import MemberNotFound, GuildNotFound


def ensure_struct(func):
    @functools.wraps(func)
    async def wrapped(*args, **kwargs):
        await Sqlite._initialize(args[0].db)
        return await func(*args, **kwargs)

    return wrapped


# noinspection SqlNoDataSourceInspection,SqlResolve
class Sqlite(Datastore):
    """A simplistic, serverless SQL implementation of the Datastore Protocol"""

    _initialized = False

    def __init__(self):
        # TODO Allow customization of storage location
        self.cwd = self._get_path()

        self.db = os.path.join(self.cwd, "datastore.db")

    async def fetch_guild(self, guild_id: int) -> LevellingGuild:
        members = await self.fetch_all_members(guild_id=guild_id)
        if not bool(members):
            # Since guilds dont have a table, its based off members
            raise GuildNotFound

        return LevellingGuild(id=guild_id, members=members)

    @ensure_struct
    async def fetch_member(
        self, member_id: int, guild_id: int = None
    ) -> LevellingMember:
        if guild_id:
            async with aiosqlite.connect(self.db) as db:
                async with db.execute(
                    "SELECT * FROM LevellingMember "
                    " WHERE guild_id=:guild_id AND id=:id",
                    {
                        "guild_id": guild_id,
                        "id": member_id,
                    },
                ) as cursor:
                    value = await cursor.fetchone()
                    if not value:
                        raise MemberNotFound

                    return LevellingMember(id=value[0], xp=value[1], guild_id=value[2])

        else:
            # No guild, so just search by id
            async with aiosqlite.connect(self.db) as db:
                async with db.execute(
                    "SELECT * FROM LevellingMember WHERE id=:identifer",
                    {
                        "id": member_id,
                    },
                ) as cursor:
                    value = await cursor.fetchone()
                    if not value:
                        raise MemberNotFound

                    return LevellingMember(id=value[0], xp=value[1])

    @ensure_struct
    async def set_member(
        self, member_id: int, data: dict, guild_id: int = None
    ) -> None:
        async with aiosqlite.connect(self.db) as db:
            await db.execute(
                "INSERT INTO LevellingMember "
                "   VALUES (:id, :xp, :guild_id) "
                "ON CONFLICT(id) "
                "   DO UPDATE "
                "   SET xp=:xp",
                {
                    "id": member_id,
                    "xp": data.get("xp", 0),
                    "guild_id": guild_id,
                },
            )
            await db.commit()

    @ensure_struct
    async def fetch_all_members(self, guild_id: int = None) -> List[LevellingMember]:
        """Used internally to populate the LevellingGuild"""
        if guild_id:
            async with aiosqlite.connect(self.db) as db:
                async with db.execute(
                    "SELECT (id, xp, guild_id) FROM LevellingMember "
                    " WHERE guild_id=:guild_id",
                    {
                        "guild_id": guild_id,
                    },
                ) as cursor:
                    values = await cursor.fetchall()
                    if not values:
                        return []

                    data = []
                    for value in values:
                        data.append(
                            LevellingMember(id=value[0], xp=value[1], guild_id=guild_id)
                        )

                    return data
        else:
            async with aiosqlite.connect(self.db) as db:
                async with db.execute("SELECT * FROM LevellingMember") as cursor:
                    values = await cursor.fetchall()
                    if not values:
                        return []

                    data = []
                    for value in values:
                        data.append(
                            LevellingMember(id=value[0], xp=value[1], guild_id=guild_id)
                        )

                    return data

    @staticmethod
    async def _initialize(db) -> None:
        """A static method used to make sure the relevant tables exist"""
        if Sqlite._initialized:
            # We are initialized
            return

        async with aiosqlite.connect(db) as db:
            async with db.execute(
                "SELECT name FROM sqlite_master WHERE type='table' AND name='LevellingMember'"
            ) as cursor:
                if not await cursor.fetchone():
                    await db.execute(
                        "CREATE TABLE LevellingMember ("
                        "   id number NOT NULL PRIMARY KEY, "
                        "   xp number NOT NULL,"
                        "   guild_id number"
                        ")"
                    )

                    await db.execute(
                        "CREATE UNIQUE INDEX id_index ON LevellingMember(id)"
                    )
                    await db.execute(
                        "CREATE UNIQUE INDEX guild_index ON LevellingMember(guild_id)"
                    )
                    await db.execute(
                        "CREATE UNIQUE INDEX shared_index ON LevellingMember(id, guild_id)"
                    )

                    await db.commit()

        Sqlite._initialized = True

    @staticmethod
    def _get_path() -> str:
        return os.path.dirname(sys.argv[0])



from typing import Optional

from attr import asdict

from levelling.abc import Datastore, Cache
from levelling.dataclass import LevellingMember
from levelling.exceptions import MemberNotFound


class Store:
    """
    An abstraction layer for storage classes,
    providing easy access to data from both caches/persistent storage.
    """

    def __init__(self, *, cache: Cache, datastore: Datastore):
        self.cache = cache
        self.datastore = datastore

    async def create_or_fetch_member(
        self, member_id: int, guild_id: int
    ) -> LevellingMember:
        """
        Fetches a member, creates if one can't be found.

        Parameters
        ----------
        member_id: int
            The id for the member
        guild_id: int
            The id for the guild of the member

        Returns
        -------
        LevellingMember
            The stored/cached/created member
        """
        member: Optional[LevellingMember] = None
        try:
            member: LevellingMember = await self.fetch_member(
                member_id=member_id, guild_id=guild_id
            )
        except MemberNotFound:
            member = LevellingMember(id=member_id, guild_id=guild_id)
        finally:
            return member

    async def fetch_member(self, member_id: int, guild_id: int) -> LevellingMember:
        """
        Fetches a member, errors if one can't be found.

        Parameters
        ----------
        member_id: int
            The id for the member
        guild_id: int
            The id for the guild of the member

        Returns
        -------
        LevellingMember
            The stored/cached member

        Raises
        ------
        MemberNotFound
            The member could not be found
        """
        try:
            member: LevellingMember = await self.cache.get_member(
                member_id=member_id, guild_id=guild_id
            )
        except MemberNotFound:
            try:
                member: LevellingMember = await self.datastore.fetch_member(
                    member_id=member_id, guild_id=guild_id
                )
            except MemberNotFound as e:
                raise e

        return member

    async def set_member(self, member: LevellingMember) -> None:
        """
        Stores a member in both cache and persistent storage.

        Parameters
        ----------
        member: LevellingMember
            The member to store
        """
        await self.datastore.set_member(
            member_id=member.id, data=asdict(member), guild_id=member.guild_id
        )
        await self.cache.set_member(
            member_id=member.id, data=asdict(member), guild_id=member.guild_id
        )



import json
import os
import sys

from levelling.abc import Datastore
from levelling.dataclass import LevellingGuild, LevellingMember
from levelling.exceptions import GuildNotFound, MemberNotFound


class Json(Datastore):
    """This class **only** exists for testing purposes. Don't use it

    Seriously its bad, its blocking, don't use it.
    """

    def __init__(self):
        self.path = self._get_path()

    async def fetch_guild(self, guild_id: int) -> LevellingGuild:
        data = await self._read()
        try:
            return LevellingGuild(id=guild_id, raw_members=data[str(guild_id)])
        except KeyError:
            raise GuildNotFound from None

    async def fetch_member(
        self, member_id: int, guild_id: int = None
    ) -> LevellingMember:
        if guild_id:
            try:
                guild = await self.fetch_guild(guild_id=guild_id)
                member: LevellingMember = next(
                    i for i in guild.members if i.id == member_id
                )
                return member
            except GuildNotFound:
                raise MemberNotFound from None
            except KeyError:
                raise MemberNotFound from None
            except StopIteration:
                raise MemberNotFound

        data = await self._read()
        try:
            return data[str(member_id)]
        except KeyError:
            raise MemberNotFound from None

    async def set_member(
        self, member_id: int, data: dict, guild_id: int = None
    ) -> None:
        file_data = await self._read()
        member_id = str(member_id)
        if guild_id:
            guild_id = str(guild_id)
            if guild_id not in file_data:
                file_data[guild_id] = {}
            file_data[guild_id][member_id] = data

        else:
            file_data[member_id] = data

        await self._write(file_data)

    async def _read(self) -> dict:
        with open(os.path.join(self.path, "store.json"), "r") as f:
            # noinspection PyTypeChecker
            return json.load(f)

    async def _write(self, data: dict) -> None:
        with open(os.path.join(self.path, "store.json"), "w") as f:
            json.dump(data, f, indent=4)

    @staticmethod
    def _get_path() -> str:
        return os.path.dirname(sys.argv[0])
